#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import bisect
import json
import numbers
import sys
import math

import numpy

import oamap.util

if sys.version_info[0] > 2:
    xrange = range

# base class of all runtime types that require proxies: List, Record, and Tuple
class Proxy(object): pass

def tojson(value):
    if isinstance(value, ListProxy):
        return [tojson(x) for x in value]
    elif isinstance(value, RecordProxy):
        return dict((n, tojson(getattr(value, n))) for n in value._fields)
    elif isinstance(value, TupleProxy):
        return [tojson(x) for x in value]
    elif isinstance(value, (numbers.Integral, numpy.integer)):
        return int(value)
    elif isinstance(value, (numbers.Real, numpy.floating)):
        if math.isnan(value):
            return "nan"
        elif value == float("-inf"):
            return "-inf"
        elif value == float("inf"):
            return "inf"
        else:
            return float(value)
    elif isinstance(value, (numbers.Complex, numpy.complex)):
        return {"real": tojson(value.real), "imag": tojson(value.imag)}
    elif isinstance(value, numpy.ndarray):
        return value.tolist()
    else:
        return value

def tojsonstring(value, *args, **kwds):
    return json.dumps(tojson(value), *args, **kwds)

def tojsonfile(file, value, *args, **kwds):
    json.dump(file, tojson(value), *args, **kwds)

################################################################ Lists

class ListProxy(Proxy):
    __slots__ = ["_generator", "_arrays", "_cache", "_whence", "_stride", "_length"]

    def __init__(self, generator, arrays, cache, whence, stride, length):
        assert stride != 0
        assert length >= 0
        self._generator = generator
        self._arrays = arrays
        self._cache = cache
        self._whence = whence
        self._stride = stride
        self._length = length

    def __repr__(self, memo=None):
        if memo is None:
            memo = set()
        key = (id(self._generator), self._whence, self._stride, self._length)
        if key in memo:
            return "[...]"
        memo = memo.union(set([key]))
        if len(self) > 10:
            before = self[:5]
            after = self[-5:]
            return "[{0}, ..., {1}]".format(", ".join(x.__repr__(memo) if isinstance(x, (ListProxy, TupleProxy)) else repr(x) for x in before),
                                            ", ".join(x.__repr__(memo) if isinstance(x, (ListProxy, TupleProxy)) else repr(x) for x in after))
        else:
            return "[{0}]".format(", ".join(x.__repr__(memo) if isinstance(x, (ListProxy, TupleProxy)) else repr(x) for x in self))

    def __str__(self):
        return repr(self)

    def __getattr__(self, field):
        if field in self.__dict__:
            return self.__dict__[field]
        else:
            import oamap.operations
            for n, x in reversed(list(oamap.operations.actions.items()) + list(oamap.operations.transformations.items()) + list(oamap.operations.recastings.items())):
                if field == n:
                    return lambda *args, **kwargs: x(self, *args, **kwargs)
            raise AttributeError("ListProxy has no attribute {0}".format(repr(field)))

    @property
    def schema(self):
        return self._generator.schema

    @property
    def fields(self):
        generator = self._generator
        while isinstance(generator, oamap.generator.ListGenerator):
            generator = generator.content
        if isinstance(generator, oamap.generator.RecordGenerator):
            return list(generator.fields)
        else:
            raise TypeError("list does not contain records")

    def indexed(self):
        return self

    def __len__(self):
        return self._length

    def __getslice__(self, start, stop):
        return self.__getitem__(slice(start, stop))

    def __getitem__(self, index):
        if isinstance(index, slice):
            start, stop, step = oamap.util.slice2sss(index, self._length)

            whence = self._whence + self._stride*start
            stride = self._stride*step

            # length = int(math.ceil(float(abs(stop - start)) / abs(step)))
            d, m = divmod(abs(start - stop), abs(step))
            length = d + (1 if m != 0 else 0)

            return ListProxy(self._generator, self._arrays, self._cache, whence, stride, length)

        else:
            normalindex = index if index >= 0 else index + self._length
            if not 0 <= normalindex < self._length:
                raise IndexError("index {0} is out of bounds for size {1}".format(index, self._length))
            return self._generator.content._generate(self._arrays, self._whence + self._stride*normalindex, self._cache)

    def __iter__(self):
        return (self._generator.content._generate(self._arrays, i, self._cache) for i in xrange(self._whence, self._whence + self._stride*self._length, self._stride))

    def __hash__(self):
        # lists aren't usually hashable, but since ListProxy is immutable, we can add this feature
        return hash((ListProxy,) + tuple(self))

    def __eq__(self, other):
        if isinstance(other, ListProxy):
            return list(self) == list(other)
        elif isinstance(other, list):
            return list(self) == other
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, ListProxy):
            return list(self) < list(other)
        elif isinstance(other, list):
            return list(self) < other
        else:
            raise TypeError("unorderable types: list() < {1}()".format(other.__class__))

    # all of the following emulate normal list functionality using the overloaded methods above

    def __ne__(self, other): return not self.__eq__(other)
    def __le__(self, other): return self.__lt__(other) or self.__eq__(other)
    def __gt__(self, other): return not self.__lt__(other) and not self.__eq__(other)
    def __ge__(self, other): return not self.__lt__(other)

    def __add__(self, other): return list(self) + list(other)
    def __mul__(self, reps): return list(self) * reps
    def __rmul__(self, reps): return reps * list(self)
    def __reversed__(self):
        if sys.version_info[0] <= 2:
            return (self[i - 1] for i in xrange(len(self), 0, -1))
        else:
            return (self[i - 1] for i in range(len(self), 0, -1))
    def count(self, value): return sum(1 for x in self if x == value)
    def index(self, value, *args):
        if len(args) == 0:
            start = 0
            stop = len(self)
        elif len(args) == 1:
            start = args[0]
            stop = len(self)
        elif len(args) == 2:
            start, stop = args
        else:
            raise TypeError("index() takes at most 3 arguments ({0} given)".format(1 + len(args)))
        for i, x in enumerate(self):
            if x == value:
                return i
        raise ValueError("{0} is not in list".format(value))

    def __contains__(self, value):
        for x in self:
            if x == value:
                return True
        return False

################################################################ Records

class RecordProxy(Proxy):
    __slots__ = ["_generator", "_arrays", "_cache", "_index"]

    def __init__(self, generator, arrays, cache, index):
        self._generator = generator
        self._arrays = arrays
        self._cache = cache
        self._index = index

    def __repr__(self):
        return "<{0} at index {1}>".format("Record" if self._generator.name is None else self._generator.name, self._index)

    def __str__(self):
        return repr(self)

    @property
    def _fields(self):
        return list(self._generator.fields)

    def __dir__(self):
        return dir(super(RecordProxy, self)) + list(str(x) for x in self._fields)

    def __getattr__(self, field):
        try:
            # actual field names get priority (they're not allowed to start with underscore)
            generator = self._generator.fields[field]
        except KeyError:
            # barring any conflicts with actual field names, "schema" and "fields" are convenient
            if field == "schema":
                return self._generator.schema
            elif field == "fields":
                return self._fields
            elif field == "name":
                return self._generator.name
            else:
                import oamap.operations
                for n, x in reversed(list(oamap.operations.actions.items()) + list(oamap.operations.transformations.items()) + list(oamap.operations.recastings.items())):
                    if field == n:
                        return lambda *args, **kwargs: x(self, *args, **kwargs)
                raise AttributeError("{0} object has no attribute {1}".format(repr("Record" if self._generator.name is None else self._generator.name), repr(field)))
        else:
            return generator._generate(self._arrays, self._index, self._cache)

    def __hash__(self):
        return hash((RecordProxy, self._generator.name) + tuple(self._generator.fields.items()))

    def __eq__(self, other):
        return isinstance(other, RecordProxy) and self._generator.name == other._generator.name and set(self._generator.fields) == set(other._generator.fields) and all(self.__getattr__(n) == other.__getattr__(n) for n in self._generator.fields)

    def __lt__(self, other):
        if isinstance(other, RecordProxy) and self._generator.name == other._generator.name and set(self._generator.fields) == set(other._generator.fields):
            return [self.__getattr__(n) for n in self._generator.fields] < [other.__getattr__(n) for n in self._generator.fields]
        else:
            raise TypeError("unorderable types: {0}() < {1}()".format("<type 'Record'>" if self._generator.name is None else "<type {0}>".format(repr(self._generator.name)), other.__class__))

    def __ne__(self, other): return not self.__eq__(other)
    def __le__(self, other): return self.__lt__(other) or self.__eq__(other)
    def __gt__(self, other): return not self.__lt__(other) and not self.__eq__(other)
    def __ge__(self, other): return not self.__lt__(other)

################################################################ Tuples

class TupleProxy(Proxy):
    __slots__ = ["_generator", "_arrays", "_cache", "_index"]

    def __init__(self, generator, arrays, cache, index):
        self._generator = generator
        self._arrays = arrays
        self._cache = cache
        self._index = index

    def __repr__(self, memo=None):
        if memo is None:
            memo = set()
        key = (self._index,) + tuple(id(x) for x in self._generator.types)
        if key in memo:
            return "(...)"
        memo = memo.union(set([key]))
        return "({0}{1})".format(", ".join(x.__repr__(memo) if isinstance(x, (ListProxy, TupleProxy)) else repr(x) for x in self), "," if len(self) == 1 else "")

    def __str__(self):
        return repr(self)

    def __getattr__(self, field):
        if field in self.__dict__:
            return self.__dict__[field]
        else:
            import oamap.operations
            for n, x in reversed(list(oamap.operations.actions.items()) + list(oamap.operations.transformations.items()) + list(oamap.operations.recastings.items())):
                if field == n:
                    return lambda *args, **kwargs: x(self, *args, **kwargs)
            raise AttributeError("TupleProxy has no attribute {0}".format(repr(field)))

    def __len__(self):
        return len(self._generator.types)

    def __getslice__(self, start, stop):
        # for old-Python compatibility
        return self.__getitem__(slice(start, stop))

    def __getitem__(self, index):
        if isinstance(index, slice):
            lenself = len(self)
            start = 0       if index.start is None else index.start
            stop  = lenself if index.stop  is None else index.stop
            step  = 1       if index.step  is None else index.step
            return tuple(self[i] for i in range(start, stop, step))

        else:
            return self._generator.types[index]._generate(self._arrays, self._index, self._cache)

    def __iter__(self):
        return (t._generate(self._arrays, self._index, self._cache) for t in self._generator.types)

    def __hash__(self):
        return hash(tuple(self))

    def __eq__(self, other):
        if isinstance(other, TupleProxy):
            return tuple(self) == tuple(other)
        elif isinstance(other, tuple):
            return tuple(self) == other
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, TupleProxy):
            return tuple(self) < tuple(other)
        elif isinstance(other, tuple):
            return tuple(self) < other
        else:
            raise TypeError("unorderable types: tuple() < {1}()".format(other.__class__))

    # all of the following emulate normal tuple functionality using the overloaded methods above

    def __ne__(self, other): return not self.__eq__(other)
    def __le__(self, other): return self.__lt__(other) or self.__eq__(other)
    def __gt__(self, other): return not self.__lt__(other) and not self.__eq__(other)
    def __ge__(self, other): return not self.__lt__(other)

    def __add__(self, other): return tuple(self) + tuple(other)
    def __mul__(self, reps): return tuple(self) * reps
    def __rmul__(self, reps): return reps * tuple(self)
    def __reversed__(self):
        return (self[i - 1] for i in range(len(self), 0, -1))
    def count(self, value): return sum(1 for x in self if x == value)
    def index(self, value, *args):
        if len(args) == 0:
            start = 0
            stop = len(self)
        elif len(args) == 1:
            start = args[0]
            stop = len(self)
        elif len(args) == 2:
            start, stop = args
        else:
            raise TypeError("index() takes at most 3 arguments ({0} given)".format(1 + len(args)))
        for i, x in enumerate(self):
            if x == value:
                return i
        raise ValueError("{0} is not in list".format(value))

    def __contains__(self, value):
        for x in self:
            if x == value:
                return True
        return False



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy

import oamap.database

class NumpyFileBackend(oamap.database.FilesystemBackend):
    def __init__(self, directory):
        super(NumpyFileBackend, self).__init__(directory, arraysuffix=".npy")

    @property
    def args(self):
        return (self._directory,)

    def tojson(self):
        return {"class": self.__class__.__module__ + "." + self.__class__.__name__,
                "directory": self._directory}

    @staticmethod
    def fromjson(obj, namespace):
        return NumpyFileBackend(obj["directory"])

    def instantiate(self, partitionid):
        return NumpyArrays(lambda name: self.fullname(partitionid, name, create=False),
                           lambda name: self.fullname(partitionid, name, create=True))

class NumpyArrays(object):
    def __init__(self, loadname, storename):
        self._loadname = loadname
        self._storename = storename

    def __getitem__(self, name):
        return numpy.load(self._loadname(name))

    def __setitem__(self, name, value):
        numpy.save(self._storename(name), value)

class NumpyFileDatabase(oamap.database.FilesystemDatabase):
    def __init__(self, directory, namespace=""):
        super(NumpyFileDatabase, self).__init__(directory, backends={namespace: NumpyFileBackend(directory)}, namespace=namespace)



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import sys

import numpy

import oamap.generator

if sys.version_info[0] > 2:
    basestring = str

class PackedSource(object):
    def __init__(self, source, suffix):
        self.source = source
        self.suffix = suffix

    def __repr__(self):
        return "{0}({1}{2})".format(self.__class__.__name__, repr(self.source), "".join(", " + repr(x) for x in self._tojsonargs()))

    def getall(self, roles):
        if hasattr(self.source, "getall"):
            return self.source.getall(roles)
        else:
            return dict((n, self.source[str(n)]) for n in roles)

    def putall(self, roles2arrays):
        if hasattr(self.source, "putall"):
            self.source.putall(roles2arrays)
        else:
            for n, x in roles2arrays.items():
                self.source[str(n)] = x

    def copy(self):
        return self.__class__(self.source, self.suffix)

    def anchor(self, source):
        if self.source is None:
            return self.__class__(source, self.suffix)
        else:
            return self.__class__(self.source.anchor(source), self.suffix)

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__ and self._tojsonargs() == other._tojsonargs()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((PackedSource, self.__class__.__name__, tuple(self._tojsonargs())))

    def tojsonfile(self, file, *args, **kwds):
        json.dump(self.tojson(), file, *args, **kwds)

    def tojsonstring(self, *args, **kwds):
        return json.dumps(self.tojson(), *args, **kwds)

    def tojson(self):
        out = []
        node = self
        while isinstance(node, PackedSource):
            args = self._tojsonargs()
            if len(args) == 0:
                out.append(self.__class__.__name__)
            else:
                out.append({self.__class__.__name__: args})
            node = node.source
        return out

    @staticmethod
    def fromjsonfile(file, *args, **kwds):
        return PackedSource.fromjson(json.load(file, *args, **kwds))

    @staticmethod
    def fromjsonstring(data, *args, **kwds):
        return PackedSource.fromjson(json.loads(data, *args, **kwds))

    @staticmethod
    def fromjson(data):
        if isinstance(data, list):
            source = None
            for datum in reversed(data):
                if isinstance(datum, basestring):
                    classname = datum
                    args = ()
                elif isinstance(datum, dict) and len(datum) == 1:
                    classname, = datum.keys()
                    args, = datum.values()
                else:
                    raise ValueError("source packings JSON must be a list of strings or {\"classname\": [args]} dicts")
                try:
                    cls = globals()[classname]
                except KeyError:
                    raise ValueError("source packing class {0} not found".format(repr(classname)))
                source = cls(source, *args)
            return source
        else:
            raise ValueError("source packings JSON must be a list of strings or {\"classname\": [args]} dicts")

################################################################ BitPackMasks

class MaskBitPack(PackedSource):
    def __init__(self, source, suffix="-bitpacked"):
        super(MaskBitPack, self).__init__(source, suffix)

    def _tojsonargs(self):
        if self.suffix == "-bitpacked":
            return []
        else:
            return [self.suffix]

    def getall(self, roles):
        others  = [n for n in roles if not isinstance(n, oamap.generator.MaskRole)]
        renamed = dict((oamap.generator.NoRole(str(n) + self.suffix, n.namespace), n) for n in roles if isinstance(n, oamap.generator.MaskRole))
        out = super(MaskBitPack, self).getall(others + list(renamed))
        for suffixedname, name in renamed.items():
            out[name] = self.unpack(out[suffixedname])
            del out[suffixedname]
        return out

    def putall(self, roles2arrays):
        out = {}
        for n, x in roles2arrays.items():
            if isinstance(n, oamap.generator.MaskRole):
                out[oamap.generator.NoRole(str(n) + self.suffix, n.namespace)] = self.pack(x)
            else:
                out[n] = x
        super(MaskBitPack, self).putall(out)

    @staticmethod
    def unpack(array):
        if not isinstance(array, numpy.ndarray):
            array = numpy.array(array, dtype=numpy.dtype(numpy.uint8))
        unmasked = numpy.unpackbits(array).view(numpy.bool_)
        mask = numpy.empty(len(unmasked), dtype=oamap.generator.Masked.maskdtype)
        mask[unmasked] = numpy.arange(unmasked.sum(), dtype=mask.dtype)
        mask[~unmasked] = oamap.generator.Masked.maskedvalue
        return mask

    @staticmethod
    def pack(array):
        if not isinstance(array, numpy.ndarray):
            array = numpy.array(array, dtype=oamap.generator.Masked.maskdtype)
        return numpy.packbits(array != oamap.generator.Masked.maskedvalue)

################################################################ RunLengthMasks

# TODO: run-length encoding for masks

################################################################ ListsAsCounts

class ListCounts(PackedSource):
    def __init__(self, source, suffix="-counts"):
        super(ListCounts, self).__init__(source, suffix)

    def _tojsonargs(self):
        if self.suffix == "-counts":
            return []
        else:
            return [self.suffix]

    def getall(self, roles):
        others  = [n for n in roles if not isinstance(n, (oamap.generator.StartsRole, oamap.generator.StopsRole))]
        renamed = dict((oamap.generator.NoRole(str(n) + self.suffix, n.namespace), n) for n in roles if isinstance(n, oamap.generator.StartsRole))
        out = super(ListCounts, self).getall(others + list(renamed))
        for suffixedname, name in renamed.items():
            out[name], out[name.stops] = self.fromcounts(out[suffixedname])
            del out[suffixedname]
        return out

    def putall(self, roles2arrays):
        out = {}
        for n, x in roles2arrays.items():
            if isinstance(n, oamap.generator.StartsRole):
                out[oamap.generator.NoRole(str(n) + self.suffix, n.namespace)] = self.tocounts(x, roles2arrays[n.stops])
            elif isinstance(n, oamap.generator.StopsRole):
                pass
            else:
                out[n] = x
        super(ListCounts, self).putall(out)

    @staticmethod
    def fromcounts(array):
        offsets = numpy.empty(len(array) + 1, dtype=oamap.generator.ListGenerator.posdtype)
        offsets[0] = 0
        offsets[1:] = numpy.cumsum(array)
        return offsets[:-1], offsets[1:]

    @staticmethod
    def tocounts(starts, stops):
        if not isinstance(starts, numpy.ndarray):
            starts = numpy.array(starts, dtype=oamap.generator.ListGenerator.posdtype)
        if not isinstance(starts, numpy.ndarray):
            stops = numpy.array(stops, dtype=oamap.generator.ListGenerator.posdtype)
        if not starts[0] == 0 or not numpy.array_equal(starts[1:], stops[:-1]):
            raise ValueError("starts and stops cannot be converted to a single counts array")
        return stops - starts

################################################################ DropUnionOffsets

class UnionDropOffsets(PackedSource):
    def __init__(self, source):
        super(DropUnionOffsets, self).__init__(source, "")

    def _tojsonargs(self):
        return []

    def getall(self, roles):
        nooffsets = [n for n in roles if not isinstance(n, oamap.generator.OffsetsRole)]
        out = super(UnionDropOffsets, self).getall(nooffsets)
        for n in roles:
            if isinstance(n, oamap.generator.TagsRole):
                out[n.offsets] = self.tags2offsets(out[n])
        return out

    def putall(self, roles2arrays):
        super(UnionDropOffsets, self).putall(dict((n, x) for n, x in roles2arrays.items() if not isinstance(n, oamap.generator.OffsetsRole)))

    @staticmethod
    def tags2offsets(tags):
        if not isinstance(tags, numpy.ndarray):
            tags = numpy.array(tags, dtype=oamap.generator.UnionGenerator.tagdtype)
        offsets = numpy.empty(len(tags), dtype=oamap.generator.UnionGenerator.offsetdtype)
        for tag in numpy.unique(tags):
            hastag = (tags == tag)
            offsets[hastag] = numpy.arange(hastag.sum(), dtype=offsets.dtype)
        return offsets

################################################################ CompressAll

# TODO: apply a named compression algorithm



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os.path

from setuptools import find_packages
from setuptools import setup

def get_version():
    g = {}
    exec(open(os.path.join("oamap", "version.py")).read(), g)
    return g["__version__"]

setup(name = "oamap",
      version = get_version(),
      packages = find_packages(exclude = ["tests"]),
      scripts = [],
      data_files = ["README.rst"],
      description = "Perform high-speed calculations on columnar data without creating intermediate objects.",
      long_description = open("README.rst").read().strip(),
      author = "Jim Pivarski (DIANA-HEP)",
      author_email = "pivarski@fnal.gov",
      maintainer = "Jim Pivarski (DIANA-HEP)",
      maintainer_email = "pivarski@fnal.gov",
      url = "https://github.com/diana-hep/oamap",
      download_url = "https://github.com/diana-hep/oamap/releases",
      license = "BSD 3-clause",
      test_suite = "tests",
      install_requires = ["numpy"],
      tests_require = ["uproot", "thriftpy", "python-snappy"],
      classifiers = [
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ],
      platforms = "Any",
      )



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy

import oamap.backend.root
import oamap.schema
import oamap.dataset
import oamap.proxy
from oamap.util import OrderedDict

def dataset(path, namespace=None, **kwargs):
    import uproot

    if namespace is None:
        namespace = "root.cmsnano({0})".format(repr(path))

    if "localsource" not in kwargs:
        kwargs["localsource"] = lambda path: uproot.source.file.FileSource(path, chunkbytes=8*1024, limitbytes=None)
    kwargs["total"] = False
    kwargs["blocking"] = True

    paths2entries = uproot.tree.numentries(path, "Events", **kwargs)
    if len(paths2entries) == 0:
        raise ValueError("path {0} matched no TTrees".format(repr(path)))

    offsets = [0]
    paths = []
    for path, numentries in paths2entries.items():
        offsets.append(offsets[-1] + numentries)
        paths.append(path)

    sch = schema(paths[0], namespace=namespace)
    doc = sch.doc
    sch.doc = None

    return oamap.dataset.Dataset("Events",
                                 sch,
                                 {namespace: oamap.backend.root.ROOTBackend(paths, "Events", namespace)},
                                 oamap.dataset.SingleThreadExecutor(),
                                 offsets,
                                 extension=None,
                                 packing=None,
                                 doc=doc,
                                 metadata={"schemafrom": paths[0]})

def proxy(path, namespace=None, extension=oamap.extension.common):
    import uproot

    if namespace is None:
        namespace = "root.cmsnano({0})".format(repr(path))

    def localsource(path):
        return uproot.source.file.FileSource(path, chunkbytes=8*1024, limitbytes=None)

    return _proxy(uproot.open(path, localsource=localsource)["Events"], namespace=namespace, extension=extension)

def _proxy(tree, namespace=None, extension=oamap.extension.common):
    if namespace is None:
        namespace = "root.cmsnano({0})".format(repr(path))

    schema = _schema(tree, namespace=namespace)
    generator = schema.generator(extension=extension)

    return oamap.proxy.ListProxy(generator, oamap.backend.root.ROOTArrays(tree, oamap.backend.root.ROOTBackend([tree._context.sourcepath], tree._context.treename, namespace)), generator._newcache(), 0, 1, tree.numentries)

def schema(path, namespace=None):
    import uproot

    if namespace is None:
        namespace = "root.cmsnano({0})".format(repr(path))

    def localsource(path):
        return uproot.source.file.FileSource(path, chunkbytes=8*1024, limitbytes=None)

    return _schema(uproot.open(path, localsource=localsource)["Events"], namespace=namespace)

def _schema(tree, namespace=None):
    if namespace is None:
        namespace = "root.cmsnano({0})".format(repr(path))

    schema = oamap.backend.root._schema(tree, namespace=namespace)

    groups = OrderedDict()
    for name in list(schema.content.keys()):
        if isinstance(schema.content[name], oamap.schema.List) and "_" in name:
            try:
                branch = tree[schema.content[name].starts]
            except KeyError:
                pass
            else:
                underscore = name.index("_")
                groupname, fieldname = name[:underscore], name[underscore + 1:]
                countbranchname = branch.countbranch.name
                if not isinstance(countbranchname, str):
                    countbranchname = countbranchname.decode("ascii")
                if groupname not in groups:
                    groups[groupname] = schema.content[groupname] = \
                        oamap.schema.List(oamap.schema.Record({}, name=groupname), starts=countbranchname, stops=countbranchname, namespace=namespace)
                assert countbranchname == schema.content[groupname].starts
                groups[groupname].content[fieldname] = schema.content[name].content
                del schema.content[name]

        elif "MET_" in name or name.startswith("LHE_") or name.startswith("Pileup_") or name.startswith("PV_"):
            underscore = name.index("_")
            groupname, fieldname = name[:underscore], name[underscore + 1:]
            if groupname not in groups:
                groups[groupname] = schema.content[groupname] = \
                    oamap.schema.Record({}, name=groupname)
            groups[groupname][fieldname] = schema.content[name]
            del schema.content[name]

    hlt = oamap.schema.Record({}, name="HLT")
    flag = oamap.schema.Record({}, name="Flag")
    for name in schema.content.keys():
        if name.startswith("HLT_"):
            hlt[name[4:]] = schema.content[name]
            del schema.content[name]
        if name.startswith("Flag_"):
            flag[name[5:]] = schema.content[name]
            del schema.content[name]

    schema.content["HLT"] = hlt
    schema.content["Flag"] = flag
    schema.content.name = "Event"
    return schema



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ast
import math
import numbers
import sys
import types
import time

import numpy

import oamap.schema
import oamap.generator
import oamap.proxy
import oamap.util
import oamap.compiler

recastings      = oamap.util.OrderedDict()
transformations = oamap.util.OrderedDict()
actions         = oamap.util.OrderedDict()

################################################################ general utilities

def _setindexes(input, output):
    if isinstance(input, oamap.proxy.ListProxy):
        if isinstance(output, oamap.proxy.ListProxy):
            output._whence, output._stride, output._length = input._whence, input._stride, input._length
        elif isinstance(output, (oamap.proxy.RecordProxy, oamap.proxy.TupleProxy)):
            output._index = input._whence

    elif isinstance(input, (oamap.proxy.RecordProxy, oamap.proxy.TupleProxy)):
        if isinstance(output, (oamap.proxy.RecordProxy, oamap.proxy.TupleProxy)):
            output._index = input._index
        elif isinstance(output, oamap.proxy.ListProxy):
            output._length = output._length - input._index
            output._whence = input._index

    else:
        raise AssertionError(type(input))

    return output
    
class _DualSource(object):
    def __init__(self, old, oldns):
        self.old = old
        self.new = {}

        i = 0
        self.namespace = None
        while self.namespace is None or self.namespace in oldns:
            self.namespace = "namespace-" + str(i)
            i += 1

        self._arraynum = 0

    def arrayname(self):
        trial = None
        while trial is None or trial in self.new:
            trial = "array-" + str(self._arraynum)
            self._arraynum += 1
        return trial

    def getall(self, roles):
        out = {}

        if hasattr(self.old, "getall"):
            out.update(self.old.getall([x for x in roles if x.namespace != self.namespace]))
        else:
            for x in roles:
                if x.namespace != self.namespace:
                    out[x] = self.old[str(x)]

        if hasattr(self.new, "getall"):
            out.update(self.new.getall([x for x in roles if x.namespace == self.namespace]))
        else:
            for x in roles:
                if x.namespace == self.namespace:
                    out[x] = self.new[str(x)]

        return out

    def put(self, schemanode, *arrays):
        if isinstance(schemanode, oamap.schema.Primitive):
            datarole = oamap.generator.DataRole(self.arrayname(), self.namespace)
            roles2arrays = {datarole: arrays[0]}
            schemanode.data = str(datarole)

        elif isinstance(schemanode, oamap.schema.List):
            startsrole = oamap.generator.StartsRole(self.arrayname(), self.namespace, None)
            stopsrole = oamap.generator.StopsRole(self.arrayname(), self.namespace, None)
            startsrole.stops = stopsrole
            stopsrole.starts = startsrole
            roles2arrays = {startsrole: arrays[0], stopsrole: arrays[1]}
            schemanode.starts = str(startsrole)
            schemanode.stops = str(stopsrole)

        elif isinstance(schemanode, oamap.schema.Union):
            tagsrole = oamap.generator.TagsRole(self.arrayname(), self.namespace, None)
            offsetsrole = oamap.generator.OffsetsRole(self.arrayname(), self.namespace, None)
            tagsrole.offsets = offsetsrole
            offsetsrole.tags = tagsrole
            roles2arrays = {tagsrole: arrays[0], offsetsrole: arrays[1]}
            schemanode.tags = str(tagsrole)
            schemanode.offsets = str(offsetsrole)

        elif isinstance(schemanode, oamap.schema.Record):
            pass

        elif isinstance(schemanode, oamap.schema.Tuple):
            pass

        elif isinstance(schemanode, oamap.schema.Pointer):
            positionsrole = oamap.generator.PositionsRole(self.arrayname(), self.namespace)
            roles2arrays = {positionsrole: arrays[0]}
            schemanode.positions = str(positionsrole)

        else:
            raise AssertionError(schemanode)

        if schemanode.nullable:
            maskrole = oamap.generator.MaskRole(self.arrayname(), self.namespace, dict(roles2arrays))
            roles2arrays[maskrole] = arrays[-1]
            schemanode.mask = str(maskrole)

        schemanode.namespace = self.namespace
        self.putall(roles2arrays)

    def putall(self, roles2arrays):
        if hasattr(self.new, "putall"):
            self.new.putall(roles2arrays)
        else:
            for n, x in roles2arrays.items():
                self.new[str(n)] = x

    def close(self):
        if hasattr(self.old, "close"):
            self.old.close()
        if hasattr(self.new, "close"):
            self.new.close()

    @property
    def backends(self):
        old = self.old
        while isinstance(old, _DualSource):
            old = old.old
        oldbackend = getattr(old, "backend", None)
        if oldbackend is None:
            return {}
        else:
            return {oldbackend.namespace: oldbackend}

    @staticmethod
    def collect(schema, arrays, namespace, prefix, delimiter):
        newarrays = {}
        def getarrays(node):
            if isinstance(node, _DualSource):
                getarrays(node.old)
                for n, x in node.new.items():
                    newarrays[node.namespace, n] = x
        getarrays(arrays)

        oldnames = {}
        def recurse(schema, memo):
            if isinstance(schema, oamap.schema.Primitive):
                if (schema.namespace, schema.data) in newarrays:
                    oldnames[id(schema)] = (schema.namespace, schema.data, schema.mask)
                    schema.namespace = namespace
                    schema.data = None
                    if schema.nullable:
                        schema.mask = None
            elif isinstance(schema, oamap.schema.List):
                if (schema.namespace, schema.starts) in newarrays:
                    oldnames[id(schema)] = (schema.namespace, schema.starts, schema.stops, schema.mask)
                    schema.namespace = namespace
                    schema.starts = None
                    schema.stops = None
                    if schema.nullable:
                        schema.mask = None
                recurse(schema.content, memo)
            elif isinstance(schema, oamap.schema.Union):
                if (schema.namespace, schema.tags) in newarrays:
                    oldnames[id(schema)] = (schema.namespace, schema.tags, schema.offsets, schema.mask)
                    schema.namespace = namespace
                    schema.tags = None
                    schema.offsets = None
                for possibility in schema.possibilities:
                    recurse(possibility, memo)
            elif isinstance(schema, (oamap.schema.Record, oamap.schema.Tuple)):
                if schema.nullable and (schema.namespace, schema.mask) in newarrays:
                    oldnames[id(schema)] = (schema.namespace, schema.mask)
                    schema.namespace = namespace
                    schema.mask = None
                for field in schema.fields.values():
                    recurse(field, memo)
            elif isinstance(schema, oamap.schema.Pointer):
                if id(schema) not in memo:
                    memo.add(id(schema))
                    if (schema.namespace, schema.positions) in newarrays:
                        oldnames[id(schema)] = (schema.namespace, schema.positions, schema.mask)
                        schema.namespace = namespace
                        schema.positions = None
                        if schema.nullable:
                            schema.mask = None
                    recurse(schema.target, memo)
            else:
                raise AssertionError(schema)

        recurse(schema, set())

        generator = schema.generator(prefix=prefix, delimiter=delimiter)

        roles2arrays = {}
        def recurse2(schema, generator, memo):
            if isinstance(generator, oamap.generator.ExtendedGenerator):
                generator = generator.generic

            r2a = {}
            if isinstance(schema, oamap.schema.Primitive):
                if id(schema) in oldnames:
                    oldns, olddata, oldmask = oldnames[id(schema)]
                    datarole = oamap.generator.DataRole(generator.data, generator.namespace)
                    r2a[datarole] = newarrays[oldns, olddata]
                    if schema.nullable:
                        maskrole = oamap.generator.MaskRole(generator.mask, generator.namespace, dict(r2a))
                        r2a[maskrole] = newarrays[oldns, oldmask]

            elif isinstance(schema, oamap.schema.List):
                if id(schema) in oldnames:
                    oldns, oldstarts, oldstops, oldmask = oldnames[id(schema)]
                    startsrole = oamap.generator.StartsRole(generator.starts, generator.namespace, None)
                    stopsrole = oamap.generator.StopsRole(generator.stops, generator.namespace, None)
                    startsrole.stops = stopsrole
                    stopsrole.starts = startsrole
                    r2a[startsrole] = newarrays[oldns, oldstarts]
                    r2a[stopsrole] = newarrays[oldns, oldstops]
                    if schema.nullable:
                        maskrole = oamap.generator.MaskRole(generator.mask, generator.namespace, dict(r2a))
                        r2a[maskrole] = newarrays[oldns, oldmask]
                recurse2(schema.content, generator.content, memo)

            elif isinstance(schema, oamap.schema.Union):
                if id(schema) in oldnames:
                    oldns, oldtags, oldoffsets, oldmask = oldnames[id(schema)]
                    tagsrole = oamap.generator.TagsRole(generator.tags, generator.namespace, None)
                    offsetsrole = oamap.generator.OffsetsRole(generator.offsets, generator.namespace, None)
                    tagsrole.offsets = offsetsrole
                    offsetsrole.tags = tagsrole
                    r2a[tagsrole] = newarrays[oldns, oldtags]
                    r2a[offsetsrole] = newarrays[oldns, oldoffsets]
                    if schema.nullable:
                        maskrole = oamap.generator.MaskRole(generator.mask, generator.namespace, dict(r2a))
                        r2a[maskrole] = newarrays[oldns, oldmask]
                for pschema, pgenerator in zip(schema.possibilities, generator.possibilities):
                    recurse2(pschema, pgenerator, memo)

            elif isinstance(schema, (oamap.schema.Record, oamap.schema.Tuple)):
                if schema.nullable and id(schema) in oldnames:
                    oldns, oldmask = oldnames[id(schema)]
                    maskrole = oamap.generator.MaskRole(generator.mask, generator.namespace, {})
                    r2a[maskrole] = newarrays[oldns, oldmask]
                for n in schema.fields:
                    recurse2(schema[n], generator.fields[n], memo)

            elif isinstance(schema, oamap.schema.Pointer):
                if id(schema) not in memo:
                    memo.add(id(schema))
                    if id(schema) in oldnames:
                        oldns, oldpositions, oldmask = oldnames[id(schema)]
                        positionsrole = oamap.generator.PositionsRole(generator.positions, generator.namespace)
                        r2a[positionsrole] = newarrays[oldns, oldpositions]
                        if schema.nullable:
                            maskrole = oamap.generator.MaskRole(generator.mask, generator.namespace, dict(r2a))
                            r2a[maskrole] = newarrays[oldns, oldmask]
                    recurse2(schema.target, generator.target, memo)

            else:
                raise AssertionError(schema)

            roles2arrays.update(r2a)

        if isinstance(schema, oamap.schema.List):
            recurse2(schema.content, generator.content, set())
        else:
            recurse2(schema, generator, set())

        return generator.namedschema(), roles2arrays

################################################################ fieldname/recordname

def fieldname(data, newname, at):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema()
        nodes = schema.path(at, parents=True)
        if len(nodes) < 2:
            raise TypeError("path {0} did not match a field in a record".format(repr(at)))

        for n, x in nodes[1].fields.items():
            if x is nodes[0]:
                oldname = n
                break

        del nodes[1][oldname]
        nodes[1][newname] = nodes[0]

        if isinstance(schema, oamap.schema.List):
            return _setindexes(data, schema(data._arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(data._arrays))
        
    else:
        raise TypeError("fieldname can only be applied to an OAMap proxy (List, Record, Tuple)")

recastings["fieldname"] = fieldname

def recordname(data, newname, at=""):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema()
        nodes = schema.path(at, parents=True)
        while isinstance(nodes[0], oamap.schema.List):
            nodes = (nodes[0].content,) + nodes
        if not isinstance(nodes[0], oamap.schema.Record):
            raise TypeError("path {0} did not match a record".format(repr(at)))

        nodes[0].name = newname
        if isinstance(schema, oamap.schema.List):
            return _setindexes(data, schema(data._arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(data._arrays))
        
    else:
        raise TypeError("fieldname can only be applied to an OAMap proxy (List, Record, Tuple)")

recastings["recordname"] = recordname

################################################################ project/keep/drop

def project(data, at):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema().project(at)
        if schema is None:
            raise TypeError("projection resulted in no schema")
        if isinstance(schema, oamap.schema.List) and isinstance(data, oamap.proxy.ListProxy):
            return _setindexes(data, schema(data._arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(data._arrays))
    else:
        raise TypeError("project can only be applied to an OAMap proxy (List, Record, Tuple)")

recastings["project"] = project

def keep(data, *paths):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema().keep(*paths)
        if schema is None:
            raise TypeError("keep operation resulted in no schema")
        if isinstance(schema, oamap.schema.List):
            return _setindexes(data, schema(data._arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(data._arrays))
    else:
        raise TypeError("keep can only be applied to an OAMap proxy (List, Record, Tuple)")

recastings["keep"] = keep

def drop(data, *paths):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema().drop(*paths)
        if schema is None:
            raise TypeError("drop operation resulted in no schema")
        if isinstance(schema, oamap.schema.List):
            return _setindexes(data, schema(data._arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(data._arrays))
    else:
        raise TypeError("drop can only be applied to an OAMap proxy (List, Record, Tuple)")

recastings["drop"] = drop

################################################################ split

def split(data, *paths):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema()
        newschema = []

        allpaths = []
        for path in paths:
            found = False
            for nodes in schema.paths(path, parents=True, allowtop=False):
                found = True
                allpaths.append(nodes)
            if not found:
                raise TypeError("path {0} did not match any field".format(repr(path)))

        for nodes in allpaths:
            if len(nodes) < 3 or not isinstance(nodes[1], oamap.schema.Record) or not isinstance(nodes[2], oamap.schema.List):
                raise TypeError("path {0} matches a field that is not in a List(Record({{field: ...}}))".format(repr(path)))

            if len(nodes) > 3 and not isinstance(nodes[3], oamap.schema.Record):
                raise TypeError("path {0} matches a field that is in something other than a Record(List(Record({{field: ...}}))):\n\n    {1}".format(repr(path), nodes[3].__repr__(indent="    ")))

            if len(nodes) == 3:
                outernode = oamap.schema.Record(oamap.util.OrderedDict())
                newschema.append(outernode)
                datanode, innernode, listnode = nodes[0], nodes[1], nodes[2]
            else:
                datanode, innernode, listnode, outernode = nodes[0], nodes[1], nodes[2], nodes[3]

            for n, x in innernode.fields.items():
                if x is datanode:
                    innername = n
                    break
            outername = None
            for n, x in outernode.fields.items():
                if x is listnode:
                    outername = n
                    break

            del innernode[innername]
            if len(innernode.fields) == 0 and outername is not None:
                del outernode[outername]

            outernode[innername] = listnode.copy(content=datanode)

        if len(newschema) > 0:
            if len(schema.content.fields) > 0:
                schema = oamap.schema.Record(oamap.util.OrderedDict([("original", schema)]))
            else:
                schema = oamap.schema.Record(oamap.util.OrderedDict())

            for ns in newschema:
                for n, x in ns.fields.items():
                    schema[n] = x

        if isinstance(schema, oamap.schema.List):
            return schema(data._arrays, numentries=len(data))
        else:
            return schema(data._arrays)

    else:
        raise TypeError("split can only be applied to an OAMap proxy (List, Record, Tuple)")

recastings["split"] = split

################################################################ merge

def merge(data, container, *paths):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema()

        allpaths = []
        for path in paths:
            found = False
            for nodes in schema.paths(path, parents=True, allowtop=False):
                found = True
                allpaths.append(nodes)
            if not found:
                raise TypeError("path {0} did not match any field".format(repr(path)))

        constructed = False
        try:
            nodes = schema.path(container, parents=True)

        except ValueError:
            try:
                slash = container.rindex("/")
            except ValueError:
                nodes = (schema,)
                tomake = container
            else:
                tofind, tomake = container[:slash], container[slash + 1:]
                nodes = schema.path(tofind, parents=True)
                container = tofind

            while isinstance(nodes[0], oamap.schema.List):
                nodes = (nodes[0].content,) + nodes
            if not isinstance(nodes[0], oamap.schema.Record):
                raise TypeError("container parent {0} is not a record".format(repr(container)))
            nodes[0][tomake] = oamap.schema.List(oamap.schema.Record({}))
            nodes = (nodes[0][tomake].content, nodes[0][tomake]) + nodes
            constructed = True

        else:
            while isinstance(nodes[0], oamap.schema.List):
                nodes = (nodes[0].content,) + nodes
            
        if len(nodes) < 2 or not isinstance(nodes[0], oamap.schema.Record) or not isinstance(nodes[1], oamap.schema.List):
            raise TypeError("container must be a List(Record(...))")
        
        containerrecord, containerlist = nodes[0], nodes[1]
        parents = nodes[2:]
        listnodes = []
        if not constructed:
            listnodes.append(containerlist)

        for nodes in allpaths:
            if len(nodes) < 2 or not isinstance(nodes[0], oamap.schema.List) or nodes[1:] != parents:
                raise TypeError("path {0} did not match a list at the same level as container {1}".format(repr(path), repr(container)))

            listnode, outernode = nodes[0], nodes[1]
            listnodes.append(listnode)

            for n, x in outernode.fields.items():
                if x is listnode:
                    outername = n
                    break

            del outernode[outername]
            containerrecord[outername] = listnode.content

        if len(listnodes) == 0:
            raise TypeError("at least one path must match schema elements")

        checknodes = [x for x in listnodes if x.starts is not None]
        assert len(checknodes) > 0

        if not all(x.namespace == checknodes[0].namespace and x.starts == checknodes[0].starts and x.stops == checknodes[0].stops for x in checknodes[1:]):
            ### RECONSIDER: without this fallback, merge is a pure recasting (like split) and is easy to decide whether to parallelize
            #
            # starts1, stops1 = data._generator.findbynames("List", checknodes[0].namespace, starts=checknodes[0].starts, stops=checknodes[0].stops)._getstartsstops(data._arrays, data._cache)
            # for x in checknodes[1:]:
            #     starts2, stops2 = data._generator.findbynames("List", x.namespace, starts=x.starts, stops=x.stops)._getstartsstops(data._arrays, data._cache)
            #     if not (starts1 is starts2 or numpy.array_equal(starts1, starts2)) and not (stops1 is stops2 or numpy.array_equal(stops1, stops2)):
            #         raise ValueError("some of the paths refer to lists of different lengths")
            raise ValueError("some of the paths refer to lists of different names")

        if constructed:
            containerlist.namespace = listnodes[0].namespace
            containerlist.starts = listnodes[0].starts
            containerlist.stops = listnodes[0].stops

        if isinstance(schema, oamap.schema.List):
            return schema(data._arrays, numentries=len(data))
        else:
            return schema(data._arrays)

    else:
        raise TypeError("merge can only be applied to an OAMap proxy (List, Record, Tuple)")

recastings["merge"] = merge

################################################################ parent

def parent(data, fieldname, at):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema()
        nodes = schema.path(at, parents=True)
        if len(nodes) < 2:
            raise TypeError("parent operation must be applied to a field of a record")

        parentnode = nodes[1]
        listnode = nodes[0]
        if not isinstance(listnode, oamap.schema.List):
            raise TypeError("parent operation must be applied to a field with list type")
        childnode = listnode.content

        if not isinstance(childnode, oamap.schema.Record):
            raise TypeError("parent operation must be applied to a field with list of records type")

        if listnode.nullable or childnode.nullable:
            raise NotImplementedError("nullable; need to merge masks")

        listgenerator = data._generator.findbynames("List", listnode.namespace, starts=listnode.starts, stops=listnode.stops)
        if listnode is schema:
            offsets = numpy.array([0, len(data)], dtype=oamap.generator.ListGenerator.posdtype)
            starts, stops = offsets[:1], offsets[-1:]
        else:
            starts, stops = listgenerator._getstartsstops(data._arrays, data._cache)

        if isinstance(parent.fill, types.FunctionType):
            try:
                import numba as nb
            except ImportError:
                pass
            else:
                parent.fill = nb.jit(nopython=True, nogil=True)(parent.fill)

        pointers = numpy.empty(stops.max() - starts.min(), dtype=oamap.generator.PointerGenerator.posdtype)
        parent.fill(starts, stops, pointers)

        childnode[fieldname] = oamap.schema.Pointer(parentnode)

        arrays = _DualSource(data._arrays, data._generator.namespaces())
        arrays.put(childnode[fieldname], pointers)

        if isinstance(schema, oamap.schema.List):
            return _setindexes(data, schema(arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(arrays))
            
def _parent_fill(starts, stops, pointers):
    for i in range(len(starts)):
        pointers[starts[i]:stops[i]] = i

parent.fill = _parent_fill
del _parent_fill

transformations["parent"] = parent

################################################################ index

def index(data, fieldname, at):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema()
        listnode = schema.path(at)
        if not isinstance(listnode, oamap.schema.List):
            raise TypeError("index operation must be applied to a field with list type")
        childnode = listnode.content

        if not isinstance(childnode, oamap.schema.Record):
            raise TypeError("index operation must be applied to a field with list of records type")

        if listnode.nullable or childnode.nullable:
            raise NotImplementedError("nullable; need to merge masks")

        listgenerator = data._generator.findbynames("List", listnode.namespace, starts=listnode.starts, stops=listnode.stops)
        if listnode is schema:
            offsets = numpy.array([0, len(data)], dtype=oamap.generator.ListGenerator.posdtype)
            starts, stops = offsets[:1], offsets[-1:]
        else:
            starts, stops = listgenerator._getstartsstops(data._arrays, data._cache)

        if isinstance(index.fill, types.FunctionType):
            try:
                import numba as nb
            except ImportError:
                pass
            else:
                index.fill = nb.jit(nopython=True, nogil=True)(index.fill)

        values = numpy.empty(stops.max() - starts.min(), dtype=numpy.int32)   # int32
        index.fill(starts, stops, values)

        childnode[fieldname] = oamap.schema.Primitive(values.dtype)

        arrays = _DualSource(data._arrays, data._generator.namespaces())
        arrays.put(childnode[fieldname], values)

        if isinstance(schema, oamap.schema.List):
            return _setindexes(data, schema(arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(arrays))
            
def _index_fill(starts, stops, pointers):
    for i in range(len(starts)):
        start = starts[i]
        stop = stops[i]
        pointers[start:stop] = numpy.arange(stop - start)

index.fill = _index_fill
del _index_fill

transformations["index"] = index

################################################################ tomask

def tomask(data, at, low, high=None):
    if isinstance(data, oamap.proxy.Proxy):
        schema = data._generator.namedschema()
        nodes = schema.path(at, parents=True)
        while isinstance(nodes[0], oamap.schema.List):
            nodes = (nodes[0].content,) + nodes
        node = nodes[0]

        arrays = _DualSource(data._arrays, data._generator.namespaces())

        if isinstance(node, oamap.schema.Primitive):
            generator = data._generator.findbynames("Primitive", node.namespace, data=node.data, mask=node.mask)

            primitive = generator._getdata(data._arrays, data._cache).copy()
            if node.nullable:
                mask = generator._getmask(data._arrays, data._cache).copy()
            else:
                node.nullable = True
                mask = numpy.arange(len(primitive), dtype=oamap.generator.Masked.maskdtype)

            if high is None:
                if math.isnan(low):
                    selection = numpy.isnan(primitive)
                else:
                    selection = (primitive == low)
            else:
                if math.isnan(low) or math.isnan(high):
                    raise ValueError("if a range is specified, neither of the endpoints can be NaN")
                selected_nans = numpy.isnan(primitive)
                primitive[selected_nans] = low
                selection = (primitive >= low)
                numpy.bitwise_and(selection, (primitive <= high), selection)
                selection[selected_nans] = False

            mask[selection] = oamap.generator.Masked.maskedvalue

            arrays.put(node, primitive, mask)

        else:
            raise NotImplementedError("tomask operation only defined on primitive fields; {0} matches:\n\n    {1}".format(repr(at), node.__repr__(indent="    ")))

        if isinstance(schema, oamap.schema.List):
            return _setindexes(data, schema(arrays, numentries=len(data)))
        else:
            return _setindexes(data, schema(arrays))

    else:
        raise TypeError("tomask can only be applied to an OAMap proxy (List, Record, Tuple)")

transformations["tomask"] = tomask

################################################################ flatten

def flatten(data, at=""):
    if (isinstance(data, oamap.proxy.ListProxy) and data._whence == 0 and data._stride == 1) or (isinstance(data, oamap.proxy.Proxy) and data._index == 0):
        schema = data._generator.namedschema()
        outernode = schema.path(at)
        if not isinstance(outernode, oamap.schema.List) or not isinstance(outernode.content, oamap.schema.List):
            raise TypeError("path {0} does not refer to a list within a list:\n\n    {1}".format(repr(at), outernode.__repr__(indent="    ")))
        innernode = outernode.content
        if outernode.nullable or innernode.nullable:
            raise NotImplementedError("nullable; need to merge masks")

        outergenerator = data._generator.findbynames("List", outernode.namespace, starts=outernode.starts, stops=outernode.stops)
        if outernode is schema:
            offsets = numpy.array([0, len(data)], dtype=oamap.generator.ListGenerator.posdtype)
            outerstarts, outerstops = offsets[:1], offsets[-1:]
        else:
            outerstarts, outerstops = outergenerator._getstartsstops(data._arrays, data._cache)
        innergenerator = data._generator.findbynames("List", innernode.namespace, starts=innernode.starts, stops=innernode.stops)
        innerstarts, innerstops = innergenerator._getstartsstops(data._arrays, data._cache)

        if not numpy.array_equal(innerstarts[1:], innerstops[:-1]):
            raise NotImplementedError("inner arrays are not contiguous: flatten would require the creation of pointers")

        starts = innerstarts[outerstarts]
        stops  = innerstops[outerstops - 1]

        outernode.content = innernode.content

        arrays = _DualSource(data._arrays, data._generator.namespaces())
        arrays.put(outernode, starts, stops)
        if isinstance(schema, oamap.schema.List) and outernode is not schema:
            return schema(arrays, numentries=len(data))
        else:
            return schema(arrays)

    else:
        raise TypeError("flatten can only be applied to a top-level OAMap proxy (List, Record, Tuple)")

transformations["flatten"] = flatten

################################################################ filter

def filter(data, fcn, args=(), at="", numba=True):
    if not isinstance(args, tuple):
        try:
            args = tuple(args)
        except TypeError:
            args = (args,)

    if (isinstance(data, oamap.proxy.ListProxy) and data._whence == 0 and data._stride == 1) or (isinstance(data, oamap.proxy.Proxy) and data._index == 0):
        schema = data._generator.namedschema()
        nodes = schema.path(at, parents=True)
        listnode = nodes[0]
        if not isinstance(listnode, oamap.schema.List):
            raise TypeError("path {0} does not refer to a list:\n\n    {1}".format(repr(at), listnode.__repr__(indent="    ")))
        if listnode.nullable:
            raise NotImplementedError("nullable; need to merge masks")

        listgenerator = data._generator.findbynames("List", listnode.namespace, starts=listnode.starts, stops=listnode.stops)
        
        if all(isinstance(x, (oamap.schema.Record, oamap.schema.Tuple)) for x in nodes[1:]):
            if listnode is schema:
                view = listgenerator(data._arrays, numentries=len(data))
            else:
                view = listgenerator(data._arrays)
        else:
            if listnode is schema:
                offsets = numpy.array([0, len(data)], dtype=oamap.generator.ListGenerator.posdtype)
                viewstarts, viewstops = offsets[:1], offsets[-1:]
            else:
                viewstarts, viewstops = listgenerator._getstartsstops(data._arrays, data._cache)
            viewschema = listgenerator.namedschema()
            viewarrays = _DualSource(data._arrays, data._generator.namespaces())
            viewoffsets = numpy.array([viewstarts.min(), viewstops.max()], dtype=oamap.generator.ListGenerator.posdtype)
            viewarrays.put(viewschema, viewoffsets[:1], viewoffsets[-1:])
            view = viewschema(viewarrays)

        fcn = oamap.util.stringfcn(fcn)
        params = fcn.__code__.co_varnames[:fcn.__code__.co_argcount]
        avoid = set(params)
        fcnname = oamap.util.varname(avoid, "fcn")
        fillname = oamap.util.varname(avoid, "fill")
        lenname = oamap.util.varname(avoid, "len")
        rangename = oamap.util.varname(avoid, "range")

        ptypes = oamap.util.paramtypes(args)
        if ptypes is not None:
            import numba as nb
            from oamap.compiler import typeof_generator
            ptypes = (typeof_generator(view._generator.content),) + ptypes
        fcn = oamap.util.trycompile(fcn, paramtypes=ptypes, numba=numba)
        rtype = oamap.util.returntype(fcn, ptypes)
        if rtype is not None:
            if rtype != nb.types.boolean:
                raise TypeError("filter function must return boolean, not {0}".format(rtype))

        if all(isinstance(x, (oamap.schema.Record, oamap.schema.Tuple)) for x in nodes[1:]):
            env = {fcnname: fcn}
            oamap.util.doexec("""
def {fill}({view}, {pointers}{params}):
    {i} = 0
    {numitems} = 0
    for {datum} in {view}:
        if {fcn}({datum}{params}):
            {pointers}[{numitems}] = {i}
            {numitems} += 1
        {i} += 1
    return {numitems}
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           pointers=oamap.util.varname(avoid, "pointers"),
           params="".join("," + x for x in params[1:]),
           i=oamap.util.varname(avoid, "i"),
           numitems=oamap.util.varname(avoid, "numitems"),
           datum=oamap.util.varname(avoid, "datum"),
           fcn=fcnname), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)

            pointers = numpy.empty(len(view), dtype=oamap.generator.PointerGenerator.posdtype)
            numitems = fill(*((view, pointers) + args))
            pointers = pointers[:numitems]
            offsets = numpy.array([0, numitems], dtype=oamap.generator.ListGenerator.posdtype)

        else:
            env = {fcnname: fcn, lenname: len, rangename: range if sys.version_info[0] > 2 else xrange}
            oamap.util.doexec("""
def {fill}({view}, {viewstarts}, {viewstops}, {stops}, {pointers}{params}):
    {numitems} = 0
    for {i} in {range}({len}({viewstarts})):
        for {j} in {range}({viewstarts}[{i}], {viewstops}[{i}]):
            {datum} = {view}[{j}]
            if {fcn}({datum}{params}):
                {pointers}[{numitems}] = {j}
                {numitems} += 1
        {stops}[{i}] = {numitems}
    return {numitems}
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           viewstarts=oamap.util.varname(avoid, "viewstarts"),
           viewstops=oamap.util.varname(avoid, "viewstops"),
           stops=oamap.util.varname(avoid, "stops"),
           pointers=oamap.util.varname(avoid, "pointers"),
           params="".join("," + x for x in params[1:]),
           numitems=oamap.util.varname(avoid, "numitems"),
           i=oamap.util.varname(avoid, "i"),
           range=rangename,
           len=lenname,
           j=oamap.util.varname(avoid, "j"),
           datum=oamap.util.varname(avoid, "datum"),
           fcn=fcnname), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)

            offsets = numpy.empty(len(viewstarts) + 1, dtype=oamap.generator.ListGenerator.posdtype)
            offsets[0] = 0
            pointers = numpy.empty(len(view), dtype=oamap.generator.PointerGenerator.posdtype)
            numitems = fill(*((view, viewstarts, viewstops, offsets[1:], pointers) + args))
            pointers = pointers[:numitems]

        listnode.content = oamap.schema.Pointer(listnode.content)

        if isinstance(listgenerator.content, oamap.generator.PointerGenerator):
            if isinstance(listgenerator.content, oamap.generator.Masked):
                raise NotImplementedError("nullable; need to merge masks")
            innerpointers = listgenerator.content._getpositions(data._arrays, data._cache)
            pointers = innerpointers[pointers]
            listnode.content.target = listnode.content.target.target

        arrays = _DualSource(data._arrays, data._generator.namespaces())
        arrays.put(listnode, offsets[:-1], offsets[1:])
        arrays.put(listnode.content, pointers)
        if isinstance(schema, oamap.schema.List) and listnode is schema:
            return schema(arrays, numentries=numitems)
        elif isinstance(schema, oamap.schema.List):
            return schema(arrays, numentries=len(data))
        else:
            return schema(arrays)

    else:
        raise TypeError("filter can only be applied to a top-level OAMap proxy (List, Record, Tuple)")

transformations["filter"] = filter

################################################################ define

def define(data, fieldname, fcn, args=(), at="", fieldtype=None, numba=True):
    if not isinstance(args, tuple):
        try:
            args = tuple(args)
        except TypeError:
            args = (args,)

    if (isinstance(data, oamap.proxy.ListProxy) and data._whence == 0 and data._stride == 1) or (isinstance(data, oamap.proxy.Proxy) and data._index == 0):
        schema = data._generator.namedschema()
        nodes = schema.path(at, parents=True)
        while isinstance(nodes[0], oamap.schema.List):
            nodes = (nodes[0].content,) + nodes
        if not isinstance(nodes[0], oamap.schema.Record):
            raise TypeError("path {0} does not refer to a record:\n\n    {1}".format(repr(at), nodes[0].__repr__(indent="    ")))

        if len(nodes) >= 2 and isinstance(nodes[1], oamap.schema.List):
            recordnode, listnode = nodes[:2]
            listgenerator = data._generator.findbynames("List", listnode.namespace, starts=listnode.starts, stops=listnode.stops)
            if listnode is schema:
                offsets = numpy.array([0, len(data)], dtype=oamap.generator.ListGenerator.posdtype)
                viewstarts, viewstops = offsets[:1], offsets[-1:]
            else:
                viewstarts, viewstops = listgenerator._getstartsstops(data._arrays, data._cache)
            viewschema = listgenerator.namedschema()
            viewarrays = _DualSource(data._arrays, data._generator.namespaces())

            if not numpy.array_equal(viewstarts[1:], viewstops[:-1]):
                raise NotImplementedError("'define' through a list defined by arrays that are not contiguous: view would require the creation of pointers")

            viewarrays.put(viewschema, viewstarts[:1], viewstops[-1:])   # unlike 'flatten', this does not preserve upper list structure (which is desirable here and not there)
            view = viewschema(viewarrays)

        else:
            recordnode = nodes[0]
            viewschema = oamap.schema.List(recordnode)
            viewarrays = _DualSource(data._arrays, data._generator.namespaces())
            offsets = numpy.array([0, 1], dtype=oamap.generator.ListGenerator.posdtype)
            viewarrays.put(viewschema, offsets[:1], offsets[-1:])
            view = viewschema(viewarrays)

        fcn = oamap.util.stringfcn(fcn)
        params = fcn.__code__.co_varnames[:fcn.__code__.co_argcount]
        avoid = set(params)
        fcnname = oamap.util.varname(avoid, "fcn")
        fillname = oamap.util.varname(avoid, "fill")

        ptypes = oamap.util.paramtypes(args)
        if ptypes is not None:
            import numba as nb
            from oamap.compiler import typeof_generator
            ptypes = (typeof_generator(view._generator.content),) + ptypes
        fcn = oamap.util.trycompile(fcn, paramtypes=ptypes, numba=numba)
        rtype = oamap.util.returntype(fcn, ptypes)

        if fieldtype is None:
            if rtype is None or rtype == nb.types.pyobject:
                fieldtype = oamap.schema.Primitive(numpy.float64, nullable=True)
            elif isinstance(rtype, (nb.types.Integer, nb.types.Float, nb.types.Boolean)):
                fieldtype = oamap.schema.Primitive(rtype.name)
            elif isinstance(rtype, nb.types.Optional) and isinstance(rtype.type, (nb.types.Integer, nb.types.Float, nb.types.Boolean)):
                fieldtype = oamap.schema.Primitive(rtype.type.name, nullable=True)
            else:
                raise NotImplementedError("'define' not implemented for type {0}".format(rtype))

        recordnode[fieldname] = fieldtype.deepcopy()

        if isinstance(fieldtype, oamap.schema.Primitive) and not fieldtype.nullable:
            env = {fcnname: fcn}
            oamap.util.doexec("""
def {fill}({view}, {primitive}{params}):
    {i} = 0
    for {datum} in {view}:
        {primitive}[{i}] = {fcn}({datum}{params})
        {i} += 1
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           primitive=oamap.util.varname(avoid, "primitive"),
           params="".join("," + x for x in params[1:]),
           i=oamap.util.varname(avoid, "i"),
           datum=oamap.util.varname(avoid, "datum"),
           fcn=fcnname), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)

            primitive = numpy.empty(len(view), dtype=fieldtype.dtype)
            fill(*((view, primitive) + args))

            arrays = _DualSource(data._arrays, data._generator.namespaces())
            arrays.put(recordnode[fieldname], primitive)
            if isinstance(schema, oamap.schema.List):
                return schema(arrays, numentries=len(data))
            else:
                return schema(arrays)

        elif isinstance(fieldtype, oamap.schema.Primitive):
            env = {fcnname: fcn}
            oamap.util.doexec("""
def {fill}({view}, {primitive}, {mask}{params}):
    {i} = 0
    {numitems} = 0
    for {datum} in {view}:
        {tmp} = {fcn}({datum}{params})
        if {tmp} is None:
            {mask}[{i}] = {maskedvalue}
        else:
            {mask}[{i}] = {numitems}
            {primitive}[{numitems}] = {tmp}
            {numitems} += 1
        {i} += 1
    return {numitems}
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           primitive=oamap.util.varname(avoid, "primitive"),
           mask=oamap.util.varname(avoid, "mask"),
           params="".join("," + x for x in params[1:]),
           i=oamap.util.varname(avoid, "i"),
           numitems=oamap.util.varname(avoid, "numitems"),
           datum=oamap.util.varname(avoid, "datum"),
           tmp=oamap.util.varname(avoid, "tmp"),
           fcn=fcnname,
           maskedvalue=oamap.generator.Masked.maskedvalue), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)

            primitive = numpy.empty(len(view), dtype=fieldtype.dtype)
            mask = numpy.empty(len(view), dtype=oamap.generator.Masked.maskdtype)
            fill(*((view, primitive, mask) + args))

            arrays = _DualSource(data._arrays, data._generator.namespaces())
            arrays.put(recordnode[fieldname], primitive, mask)
            if isinstance(schema, oamap.schema.List):
                return schema(arrays, numentries=len(data))
            else:
                return schema(arrays)

        else:
            raise NotImplementedError("define not implemented for fieldtype:\n\n    {0}".format(fieldtype.__repr__(indent="    ")))

    else:
        raise TypeError("define can only be applied to a top-level OAMap proxy (List, Record, Tuple)")

transformations["define"] = define

################################################################ map

def map(data, fcn, args=(), at="", names=None, numba=True):
    if not isinstance(args, tuple):
        try:
            args = tuple(args)
        except TypeError:
            args = (args,)

    if (isinstance(data, oamap.proxy.ListProxy) and data._whence == 0 and data._stride == 1) or (isinstance(data, oamap.proxy.Proxy) and data._index == 0):
        listnode = data._generator.namedschema().path(at)
        if not isinstance(listnode, oamap.schema.List):
            raise TypeError("path {0} does not refer to a list:\n\n    {1}".format(repr(at), listnode.__repr__(indent="    ")))
        if listnode.nullable:
            raise NotImplementedError("nullable; need to merge masks")

        listgenerator = data._generator.findbynames("List", listnode.namespace, starts=listnode.starts, stops=listnode.stops)

        viewstarts, viewstops = listgenerator._getstartsstops(data._arrays, data._cache)
        viewschema = listgenerator.namedschema()
        viewarrays = _DualSource(data._arrays, data._generator.namespaces())
        viewoffsets = numpy.array([viewstarts.min(), viewstops.max()], dtype=oamap.generator.ListGenerator.posdtype)
        viewarrays.put(viewschema, viewoffsets[:1], viewoffsets[-1:])
        view = viewschema(viewarrays)

        fcn = oamap.util.stringfcn(fcn)
        params = fcn.__code__.co_varnames[:fcn.__code__.co_argcount]
        avoid = set(params)
        fcnname = oamap.util.varname(avoid, "fcn")
        fillname = oamap.util.varname(avoid, "fill")

        ptypes = oamap.util.paramtypes(args)
        if ptypes is not None:
            import numba as nb
            from oamap.compiler import typeof_generator
            ptypes = (typeof_generator(view._generator.content),) + ptypes
        fcn = oamap.util.trycompile(fcn, paramtypes=ptypes, numba=numba)
        rtype = oamap.util.returntype(fcn, ptypes)

        if rtype is None:
            viewindex = 0
            for datum in view:
                first = fcn(*((datum,) + args))
                viewindex += 1
                if first is not None:
                    break

            if viewindex == len(view):
                out = None

            else:
                if isinstance(first, (numbers.Integral, numbers.Real, numpy.integer, numpy.floating, bool, numpy.bool_)):
                    out = numpy.empty(len(view), dtype=numpy.float64)

                elif isinstance(first, tuple) and len(first) > 0 and all(isinstance(x, (numbers.Integral, numbers.Real, numpy.integer, numpy.floating, bool, numpy.bool_)) for x in first):
                    if names is None:
                        if hasattr(first, "_fields"):
                            names = first._fields
                        else:
                            names = ["f" + str(i) for i in range(len(first))]
                    if len(names) != len(first):
                        raise TypeError("names has length {0} but function returns {1} numbers per row".format(len(names), len(first)))

                    out = numpy.empty(len(view), dtype=list(zip(names, [numpy.float64] * len(first))))

                else:
                    print()
                    print(isinstance(first, tuple))
                    print(len(first))
                    print([isinstance(x, (numbers.Integral, numbers.Real, numpy.integer, numpy.floating, bool, numpy.bool_)) for x in first])

                    raise TypeError("function must return tuples of numbers (rows of a table)")

                numitems = 0
                out[numitems] = first
                numitems += 1
                if args == ():
                    for datum in view[viewindex:]:
                        tmp = fcn(datum)
                        if tmp is not None:
                            out[numitems] = tmp
                            numitems += 1
                else:
                    for datum in view[viewindex:]:
                        tmp = fcn(*((datum,) + args))
                        if tmp is not None:
                            out[numitems] = tmp
                            numitems += 1

                out = out[:numitems]
                        
        elif isinstance(rtype, (nb.types.Integer, nb.types.Float, nb.types.Boolean)):
            out = numpy.empty(len(view), dtype=numpy.dtype(rtype.name))
            env = {fcnname: fcn}
            oamap.util.doexec("""
def {fill}({view}, {out}{params}):
    {numitems} = 0
    for {datum} in {view}:
        {out}[{numitems}] = {fcn}({datum}{params})
        {numitems} += 1
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           out=oamap.util.varname(avoid, "out"),
           params="".join("," + x for x in params[1:]),
           numitems=oamap.util.varname(avoid, "numitems"),
           datum=oamap.util.varname(avoid, "datum"),
           fcn=fcnname), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)
            fill(*((view, out) + args))

        elif isinstance(rtype, nb.types.Optional) and isinstance(rtype.type, (nb.types.Integer, nb.types.Float, nb.types.Boolean)):
            out = numpy.empty(len(view), dtype=numpy.dtype(rtype.type.name))
            env = {fcnname: fcn}
            oamap.util.doexec("""
def {fill}({view}, {out}{params}):
    {numitems} = 0
    for {datum} in {view}:
        {tmp} = {fcn}({datum}{params})
        if {tmp} is not None:
            {out}[{numitems}] = {tmp}
            {numitems} += 1
    return {numitems}
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           out=oamap.util.varname(avoid, "out"),
           params="".join("," + x for x in params[1:]),
           numitems=oamap.util.varname(avoid, "numitems"),
           datum=oamap.util.varname(avoid, "datum"),
           tmp=oamap.util.varname(avoid, "tmp"),
           fcn=fcnname), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)
            numitems = fill(*((view, out) + args))
            out = out[:numitems]

        elif isinstance(rtype, (nb.types.Tuple, nb.types.NamedTuple, nb.types.UniTuple, nb.types.NamedUniTuple)) and len(rtype.types) > 0 and all(isinstance(x, (nb.types.Integer, nb.types.Float, nb.types.Boolean)) for x in rtype.types):
            if names is None:
                if isinstance(rtype, (nb.types.NamedTuple, nb.types.NamedUniTuple)):
                    names = rtype.fields
                else:
                    names = ["f" + str(i) for i in range(len(rtype.types))]
            if len(names) != len(rtype.types):
                raise TypeError("names has length {0} but function returns {1} numbers per row".format(len(names), len(rtype.types)))

            out = numpy.empty(len(view), dtype=list(zip(names, [numpy.dtype(x.name) for x in rtype.types])))
            outs = tuple(out[n] for n in names)

            outnames = [oamap.util.varname(avoid, "out" + str(i)) for i in range(len(names))]
            numitemsname = oamap.util.varname(avoid, "numitems")
            tmpname = oamap.util.varname(avoid, "tmp")
            env = {fcnname: fcn}
            oamap.util.doexec("""
def {fill}({view}, {outs}{params}):
    {numitems} = 0
    for {datum} in {view}:
        {tmp} = {fcn}({datum}{params})
        {assignments}
        {numitems} += 1
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           outs=",".join(outnames),
           params="".join("," + x for x in params[1:]),
           numitems=numitemsname,
           datum=oamap.util.varname(avoid, "datum"),
           tmp=tmpname,
           fcn=fcnname,
           assignments="\n        ".join("{out}[{numitems}] = {tmp}[{i}]".format(out=out, numitems=numitemsname, tmp=tmpname, i=i) for i, out in enumerate(outnames))), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)
            fill(*((view,) + outs + args))

        elif isinstance(rtype, nb.types.Optional) and isinstance(rtype.type, (nb.types.Tuple, nb.types.NamedTuple, nb.types.UniTuple, nb.types.NamedUniTuple)) and len(rtype.type.types) > 0 and all(isinstance(x, (nb.types.Integer, nb.types.Float, nb.types.Boolean)) for x in rtype.type.types):
            if names is None:
                if isinstance(rtype.type, (nb.types.NamedTuple, nb.types.NamedUniTuple)):
                    names = rtype.type.fields
                else:
                    names = ["f" + str(i) for i in range(len(rtype.type.types))]
            if len(names) != len(rtype.type.types):
                raise TypeError("names has length {0} but function returns {1} numbers per row".format(len(names), len(rtype.type.types)))

            out = numpy.empty(len(view), dtype=list(zip(names, [numpy.dtype(x.name) for x in rtype.type.types])))
            outs = tuple(out[n] for n in names)

            outnames = [oamap.util.varname(avoid, "out" + str(i)) for i in range(len(names))]
            numitemsname = oamap.util.varname(avoid, "numitems")
            tmp2name = oamap.util.varname(avoid, "tmp2")
            requiredname = oamap.util.varname(avoid, "required")
            env = {fcnname: fcn, requiredname: oamap.compiler.required}
            oamap.util.doexec("""
def {fill}({view}, {outs}{params}):
    {numitems} = 0
    for {datum} in {view}:
        {tmp} = {fcn}({datum}{params})
        if {tmp} is not None:
            {tmp2} = {required}({tmp})
            {assignments}
            {numitems} += 1
    return {numitems}
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           outs=",".join(outnames),
           params="".join("," + x for x in params[1:]),
           numitems=numitemsname,
           datum=oamap.util.varname(avoid, "datum"),
           tmp=oamap.util.varname(avoid, "tmp"),
           tmp2=tmp2name,
           required=requiredname,
           fcn=fcnname,
           assignments="\n            ".join("{out}[{numitems}] = {tmp2}[{i}]".format(out=out, numitems=numitemsname, tmp2=tmp2name, i=i) for i, out in enumerate(outnames))), env)
            fill = oamap.util.trycompile(env[fillname], numba=numba)
            numitems = fill(*((view,) + outs + args))
            out = out[:numitems]

        else:
            raise TypeError("function must return tuples of numbers (rows of a table)")

        return out

    else:
        raise TypeError("map can only be applied to a top-level OAMap proxy (List, Record, Tuple)")

class MapCombiner(object):
    def __init__(self, futures):
        self._futures = futures
        self._result = None
    def result(self, timeout=None):
        if self._result is None:
            starttime = time.time()
            results = []
            for future in self._futures:
                if timeout is not None:
                    timeout = max(1e-6, timeout - (time.time() - starttime))
                results.append(future.result(timeout))
            self._result = numpy.concatenate(results)
        return self._result
    def done(self):
        return all(x.done() for x in self._futures)
    def exception(self, timeout=None):
        raise NotImplementedError
    def traceback(self, timeout=None):
        raise NotImplementedError

map.combiner = MapCombiner
del MapCombiner

actions["map"] = map

################################################################ reduce

def reduce(data, tally, fcn, args=(), at="", numba=True):
    if not isinstance(args, tuple):
        try:
            args = tuple(args)
        except TypeError:
            args = (args,)

    if (isinstance(data, oamap.proxy.ListProxy) and data._whence == 0 and data._stride == 1) or (isinstance(data, oamap.proxy.Proxy) and data._index == 0):
        listnode = data._generator.namedschema().path(at)
        if not isinstance(listnode, oamap.schema.List):
            raise TypeError("path {0} does not refer to a list:\n\n    {1}".format(repr(at), listnode.__repr__(indent="    ")))
        if listnode.nullable:
            raise NotImplementedError("nullable; need to merge masks")

        listgenerator = data._generator.findbynames("List", listnode.namespace, starts=listnode.starts, stops=listnode.stops)
        viewstarts, viewstops = listgenerator._getstartsstops(data._arrays, data._cache)
        viewschema = listgenerator.namedschema()
        viewarrays = _DualSource(data._arrays, data._generator.namespaces())
        viewoffsets = numpy.array([viewstarts.min(), viewstops.max()], dtype=oamap.generator.ListGenerator.posdtype)
        viewarrays.put(viewschema, viewoffsets[:1], viewoffsets[-1:])
        view = viewschema(viewarrays)

        fcn = oamap.util.stringfcn(fcn)
        if fcn.__code__.co_argcount < 2:
            raise TypeError("function must have at least two parameters (data and tally)")

        params = fcn.__code__.co_varnames[:fcn.__code__.co_argcount]
        avoid = set(params)
        fcnname = oamap.util.varname(avoid, "fcn")
        fillname = oamap.util.varname(avoid, "fill")
        tallyname = params[1]

        ptypes = oamap.util.paramtypes(args)
        if ptypes is not None:
            import numba as nb
            from oamap.compiler import typeof_generator
            ptypes = (typeof_generator(view._generator.content), nb.typeof(tally)) + ptypes
        fcn = oamap.util.trycompile(fcn, paramtypes=ptypes, numba=numba)
        rtype = oamap.util.returntype(fcn, ptypes)

        if rtype is not None:
            if nb.typeof(tally) != rtype:
                raise TypeError("function should return the same type as tally")

        env = {fcnname: fcn}
        oamap.util.doexec("""
def {fill}({view}, {tally}{params}):
    for {datum} in {view}:
        {tally} = {fcn}({datum}, {tally}{params})
    return {tally}
""".format(fill=fillname,
           view=oamap.util.varname(avoid, "view"),
           tally=tallyname,
           params="".join("," + x for x in params[2:]),
           datum=oamap.util.varname(avoid, "datum"),
           fcn=fcnname), env)
        fill = oamap.util.trycompile(env[fillname], numba=numba)

        return fill(*((view, tally) + args))

    else:
        raise TypeError("reduce can only be applied to a top-level OAMap proxy (List, Record, Tuple)")

class ReduceCombiner(object):
    def __init__(self, futures):
        self._futures = futures
        self._result = None
    def result(self, timeout=None):
        if self._result is None:
            starttime = time.time()
            result = None
            for future in self._futures:
                if timeout is not None:
                    timeout = max(1e-6, timeout - (time.time() - starttime))
                if result is None:
                    result = future.result(timeout)
                else:
                    result = result + future.result(timeout)
            self._result = result
        return self._result
    def done(self):
        return all(x.done() for x in self._futures)
    def exception(self, timeout=None):
        raise NotImplementedError
    def traceback(self, timeout=None):
        raise NotImplementedError

reduce.combiner = ReduceCombiner
del ReduceCombiner

actions["reduce"] = reduce
